/**
  *  \file game/tables.cc
  *  \brief Built-in Integer-to-String Mappings
  *
  *  This module contains integer-to-string mappings not taken from
  *  definition files.
  *
  *  (c) 2001-2015 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/tables.h"
#include "cpluslib/misc.h"      // countof
#include "game/planet.h"        // ind_XXX
#include "util/nls.h"

/** Get name for a temperature.
    \param nr government number
    \return statically-allocated, translated string. */
const char*
getTemperatureName(int temp)
{
    if (temp <= 14)
        return _("arctic");
    else if (temp <= 39)
        return _("cool");
    else if (temp <= 64)
        return _("warm");
    else if (temp <= 84)
        return _("tropical");
    else
        return _("desert");
}

/** Get name for a native race.
    \param nr government number
    \return statically-allocated, translated string. */
const char*
getNativeRaceName(int race)
{
    static const char*const names[] = {
        N_("none"),
        N_("Humanoid"),
        N_("Bovinoid"),
        N_("Reptilian"),
        N_("Avian"),
        N_("Amorphous"),
        N_("Insectoid"),
        N_("Amphibian"),
        N_("Ghipsoldal"),
        N_("Siliconoid"),
        N_("Divine"),           // This and the following are proposed native races.
        N_("Artificial"),
        N_("Spirits"),
        N_("Viral"),
        N_("Plasmaoid"),
        N_("Gaseous"),
    };
    if (race < 0 || race > int(countof(names)))
        return "???";
    else
        return _(names[race]);
}

/** Get name for a native government.
    \param nr government number
    \return statically-allocated, translated string. */
const char*
getNativeGovernmentName(int gov)
{
    static const char*const names[] = {
        N_("none"),
        N_("Anarchy"),
        N_("Pre-Tribal"),
        N_("Early-Tribal"),
        N_("Tribal"),
        N_("Feudal"),
        N_("Monarchy"),
        N_("Representative"),
        N_("Participatory"),
        N_("Unity")
    };
    if (gov < 0 || gov > int(countof(names)))
        return "???";
    else
        return _(names[gov]);
}

/** Get population happiness string.
    \param happy happiness value
    \return statically-allocated, translated string. */
const char*
getHappinessName(int happy)
{
    if (happy >= 90)
        return _("happy");
    else if (happy >= 70)
        return _("calm");
    else if (happy >= 50)
        return _("unhappy");
    else if (happy >= 40)
        return _("very angry");
    else if (happy >= 20)
        return _("rioting");
    else
        return _("fighting");
}

/** Get population happiness change string.
    \param change happiness change
    \return statically-allocated, translated string. */
const char*
getHappyChangeName(int change)
{
    if (change < -5)
        return _("They HATE you!");
    else if (change < 0)
        return _("They are angry about you!");
    else if (change == 0)
        return _("They are undecided about you.");
    else if (change > 4)
        return _("They LOVE you.");
    else
        return _("They like your leadership.");
}

/** Get starbase mission name.
    \param mission mission number
    \return statically-allocated, translated string. */
const char*
getBaseMissionName(int mission)
{
    switch (mission) {
     case 0:
        return _("none");
     case 1:
        return _("Refuel ships");
     case 2:
        return _("Maximize defense");
     case 3:
        return _("Load torpedoes onto ships");
     case 4:
        return _("Unload incoming ships");
     case 5:
        return _("Repair base");
     case 6:
        return _("Force surrender");
     default:
        return "???";
    }
}

/** Get industry level name.
    \param level level, one of the ind_XXX constants.
    We do not insist on getting an ind_XXX constant, and do the
    right thing for parameters outside that range.
    \return statically-allocated, translated string */
const char*
getIndustryLevelName(int level)
{
    if (level <= ind_Minimal)
        return _("minimal");
    else if (level <= ind_Light)
        return _("light");
    else if (level <= ind_Moderate)
        return _("moderate");
    else if (level <= ind_Substantial)
        return _("substantial");
    else
        return _("heavy");
}

/** Get mineral mass class name.
    \param mass Amount of core mineral
    \return statically-allocated, translated string */
const char*
getMineralMassClassName(int32_t mass)
{
    if (mass >= 5000)
        return _("abundant");
    else if (mass >= 1200)
        return _("very common");
    else if (mass >= 600)
        return _("common");
    else if (mass >= 100)
        return _("rare");
    else if (mass > 0)
        return _("very rare");
    else
        return _("none");
}

/** Get mineral density class name.
    \param density Density of core mineral, percentage
    \return statically-allocated, translated string */
const char*
getMineralDensityClassName(int density)
{
    if (density < 10)
        return _("very scattered");
    else if (density < 30)
        return _("scattered");
    else if (density < 40)
        return _("dispersed");
    else if (density < 70)
        return _("concentrated");
    else
        return _("large masses");
}

/** Get name for wormhole stability code.
    \param code Stability code, [0,5]
    \return statically-allocated, translated string */
const char*
getWormholeStabilityName(int code)
{
    if (code <= 0)
        return _("very stable (<5%)");
    else if (code <= 1)
        return _("stable (<15%)");
    else if (code <= 2)
        return _("mostly stable (<30%)");
    else if (code <= 3)
        return _("unstable (<50%)");
    else if (code <= 4)
        return _("very unstable (<80%)");
    else
        return _("completely unstable");
}

/** Get name for a heading (direction).
    \param heading Heading in degrees
    \return statically-allocated, translated string */
const char*
getHeadingName(int heading)
{
    // FIXME: those cannot easily be translated because they're so short.
    // One way to make them translatable is to put them into a big string
    // containing all headings, and have this function return a string.
    static const char names[][4] = {
        "N", "NNE", "NE", "ENE",
        "E", "ESE", "SE", "SSE",
        "S", "SSW", "SW", "WSW",
        "W", "WNW", "NW", "NNW"
    };
    if (heading < 0) {
        return _("unknown");
    } else {
        /* There are 16 headings in 360 degrees, each covering 22.5 degrees,
           and aligned to 11.25 degrees. We can remove the fractional digits
           easily by multiplying by four. */
        return names[((4*heading + 45) / 90) & 15];
    }
}

/** Get name for an ion storm class.
    \param voltage Class of storm
    \return statically-allocated, translated string */
const char*
getIonStormClassName(int voltage)
{
    if (voltage < 50)
        return _("harmless");
    else if (voltage < 100)
        return _("moderate");
    else if (voltage < 150)
        return _("strong");
    else if (voltage < 200)
        return _("dangerous");
    else
        return _("VERY dangerous");
}

/** Get name for a planetary structure.
    Note that unlike the other items in this file, this is a PCC2-internal convention,
    not universal to the data formats or other clients.
    The return value always is a plural form usable for a heading.
    \param kind Structure number (str_XXX)
    \return statically-allocated, translated string */
const char*
getPlanetaryStructureName(int kind)
{
    CTASSERT(str_Mines == 0);
    CTASSERT(str_Factories == 1);
    CTASSERT(str_Defense == 2);
    CTASSERT(str_BaseDefense == 3);
    static const char*const names[] = {
        N_("Mineral Mines"),
        N_("Factories"),
        N_("Defense Posts"),
        N_("Starbase Defense")
    };
    if (kind >= 0 && kind < int(countof(names))) {
        return _(names[kind]);
    } else {
        return "???";
    }
}
