/**
  *  \file game/unpack.h
  *  \brief Unpacker core
  *
  *  (c) 2001-2015 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_UNPACK_H
#define PCC_V2_GAME_UNPACK_H

#include "cpluslib/types.h"
#include "game/gen.h"
#include "game/sweep.h"
#include "cpluslib/buffer.h"
#include "util/answer.h"

class GResultFile;
class Directory;
class AnswerProvider;
class ProgressMonitor;
class GUnpackModifier;

class GUnpacker : public GGameDirManipulator {
 public:
    enum Format {
        WindowsFormat,
        DosFormat
    };
    enum {
        qAcceptRaceNames
    };

    GUnpacker();
    ~GUnpacker();

    void setFormat(Format fmt);
    void setIgnore35Part(bool flag);
    void setCreateTargetExt(bool flag);
    void setFixErrors(bool flag);
    void setForceIgnoreErrors(bool flag);
    void setAnswerProvider(AnswerProvider& ap);

    void unpack(GResultFile& result, int player, ProgressMonitor& pm, GUnpackModifier* modif);

    int getTurnNumber() const;

    Format getFormat() const;

 private:
    Format opt_format;          ///< Format of directory to create.
    bool   opt_ignore_35;       ///< true to ignore 3.5 part of RST.
    bool   opt_target_ext;      ///< true to generate TARGETx.EXT file.
    bool   opt_errors;          ///< true to fix some errors.
    bool   opt_force;           ///< true to unpack despite checksum errors.

    AnswerProvider* opt_answer; ///< Answer provider.

    GGen   gen;
    Buffer<char> target_ext_buffer;

    void unpackShips(GResultFile& result, int player, ProgressMonitor& pm, GUnpackModifier* modif);
    void unpackPlanets(GResultFile& result, int player, ProgressMonitor& pm, GUnpackModifier* modif);
    void unpackBases(GResultFile& result, int player, ProgressMonitor& pm, GUnpackModifier* modif);
    void unpackTargets(GResultFile& result, int player, ProgressMonitor& pm);
    void unpackVCRs(GResultFile& result, int player, ProgressMonitor& pm);
    void unpackShipXY(GResultFile& result, int player, int32_t space, ProgressMonitor& pm);
    void unpackMessages(GResultFile& result, int player, ProgressMonitor& pm);
    void unpackKore(GResultFile& result, int player, ProgressMonitor& pm);
    void unpackSkore(GResultFile& result, int player, ProgressMonitor& pm);

    void createBlankFiles(int player);

    void saveGen(int player);
    void saveTargetExt(int player);

    void copyAndFix(TShip& dat, TShip& dis);
    void copyAndFix(TPlanet& dat, TPlanet& dis);
    void copyAndFix(TStarbase& dat, TStarbase& dis);

    void updateIndex(int player);
};

/** Set file format to produce. This defines what format the checksums
    are generated in, it does not prevent the unpacker from extracting
    a possible Windows part of the RST (see setIgnore35Part). */
inline void
GUnpacker::setFormat(Format fmt)
{
    opt_format = fmt;
}

/** Set version 3.5 part handling. If this option is set, a version
    3.5 part of the RST will be ignored as if it were not present. */
inline void
GUnpacker::setIgnore35Part(bool flag)
{
    opt_ignore_35 = flag;
}

/** Set target.ext creation flag. If this option is set, a target.ext
    file will be made containing all targets beyond the 50th from the
    regular target section, plus all AVCs from the Winplan part of the
    RST. */
inline void
GUnpacker::setCreateTargetExt(bool flag)
{
    opt_target_ext = flag;
}

/** Set error correction flag. If enabled, a number of host-side
    errors will be corrected to make the produced game directory a
    little more compatible. */
inline void
GUnpacker::setFixErrors(bool flag)
{
    opt_errors = flag;
}

/** Set checksum ignore flag. If enabled, checksum errors will be
    ignored and not cause the unpack process to be aborted. */
inline void
GUnpacker::setForceIgnoreErrors(bool flag)
{
    opt_force = flag;
}

/** Set answer provider. It will be asked using qAcceptRaceNames etc. */
inline void
GUnpacker::setAnswerProvider(AnswerProvider& ap)
{
    opt_answer = &ap;
}

/** Get turn number. Usable after successful unpacking. */
inline int
GUnpacker::getTurnNumber() const
{
    return gen.getTurnNumber();
}

/** Get configured format. */
inline GUnpacker::Format
GUnpacker::getFormat() const
{
    return opt_format;
}

#endif
