/**
  *  \file int/exec.h
  *  \brief Interpreter: Bytecode Process Execution
  *
  *  (c) 2001-2015 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_INT_EXEC_H
#define PCC_V2_INT_EXEC_H

#include <memory>
#include "cpluslib/smartptr.h"
#include "int/bytecode.h"
#include "cpluslib/ptrvector.h"
#include "int/context.h"
#include "int/interror.h"

class IntVMSaveContext;
class IntVMLoadContext;
class IntNotificationMessage;

/** Stack frame of an executing process. */
struct IntExecutionFrame {
    /** Bytecode object executing in this frame. */
    IntBCORef bco;

    /** Next instruction to execute. */
    IntBytecodeObject::pc_t pc;

    /** Local values (parameters, local variables). */
    IntDataSegment local_values;

    /** Local names (parameters, local variables). */
    IntVariableNames local_names;

    /** Top of context stack when this frame was opened. */
    uint32_t context_sp;

    /** Top of exception stack when this frame was opened. */
    uint32_t exception_sp;

    /** Own index. */
    uint32_t frame_sp;

    /** Set if caller wants a result on the stack. That is, when this frame is removed, an additional
        value must be pushed to the value stack. */
    bool want_result;

    IntExecutionFrame();
    IntExecutionFrame(IntBCORef bco);
    ~IntExecutionFrame();
};

/** Exception handler. */
struct IntExceptionHandler {
    uint32_t                frame_sp;
    uint32_t                context_sp;
    uint32_t                value_sp;
    IntBytecodeObject::pc_t pc;
};

/** Execution Context (process). */
class IntExecutionContext {
    friend class IntVMSaveContext;
    friend class IntVMLoadContext;
 public:
    enum ProcessState {
        psRunning,              ///< Process is currently running.
        psRunnable,             ///< Process wants to run.
        psTerminated,           ///< Process doesn't ever want to run again.
        psEnded,                ///< Process ran until its end.
        psSuspended,            ///< Process wants to run later.
        psError,                ///< Process terminated with an error.
        psFrozen,               ///< Process is frozen and must not be modified/run.
        psTerminating           ///< Process is about to terminate (but can be revived).
        // Additional states in PCC1: Break (Ctrl-C), Callback, Frozen
    };
    enum ProcessKind {
        pkDefault,
        pkShipTask,
        pkPlanetTask,
        pkBaseTask
    };
    
    IntExecutionContext(string_t name);
    ~IntExecutionContext();

    void pushNewFrame(IntExecutionFrame* frame);
    void popFrame();

    void pushExceptionHandler(IntBytecodeObject::pc_t pc);
    void popExceptionHandler();
    void handleException(string_t e, string_t trace);

    void pushNewContext(IntContext* ctx);
    void pushContextsFrom(ptr_vector<IntContext>& ctxs);
    void markContextTOS();
    void popContext();

    void pushNewValue(IntValue* v)
        { value_stack.pushNew(v); }
    void dropValue()
        { value_stack.pop(); }

    void            setState(ProcessState ps);
    ProcessState    getState() const;
    string_t        getName() const;
    void            setPriority(int pri);
    int             getPriority() const;
    const IntError& getError() const;
    void            setProcessKind(ProcessKind k);
    ProcessKind     getProcessKind() const;

    uint32_t getNumActiveFrames() const
        { return frames.size(); }
    IntExecutionFrame* getOutermostFrame()
        { return frames[0]; }

    IntValue* getResult()
        { return value_stack.top(); }

    uint32_t getStackSize() const
        { return value_stack.size(); }

    void addTraceTo(IntError& err);

    void run();
    void executeInstruction();
    void suspendForUI();

    IntContext* lookup(const string_t& name, int32_t& index);

    static void setNewGlobalContext(IntContext* p);
    static void setBreakHandler(bool check_break());

    bool handleLoad(string_t name);
    bool setVariable(string_t name, IntValue* value);
    IntValue* getVariable(string_t name);
    GObject* getCurrentObject() const;
    GObject* getInvokingObject() const;

    void setNotificationMessage(IntNotificationMessage* msg);
    IntNotificationMessage* getNotificationMessage() const;

    bool checkForBreak();
    void handleFunctionCall(IntBCORef bco, IntDataSegment& args, bool want_result);

    const ptr_vector<IntContext>& getContexts()
        { return contexts; }

 private:
    void checkStack(uint32_t required);
    void handleInvalidOpcode();
    void handleDim(IntDataSegment& values, IntVariableNames& names, uint16_t index);
    void handleEvalStatement(uint32_t nargs);
    void handleEvalExpression();
    void handleAddHook();
    void handleNewArray(uint16_t ndim);
    void handleResizeArray(uint16_t ndim);
    void handleMakeList(uint16_t nelems);
    void handleNewHash();
    void handleBind(uint16_t nargs);
    bool handleDecrement();
    IntValue* getReferencedValue(const IntOpcode& op);

    /** Status. */
    ProcessState process_state;

    /** Process name. */
    string_t process_name;

    /** Priority. */
    int process_priority;

    /** Error message if terminated by error. */
    IntError process_error;

    /** All active stack frames. */
    ptr_vector<IntExecutionFrame> frames;

    /** All active contexts for name lookup. */
    ptr_vector<IntContext> contexts;

    /** All active exception handling contexts. */
    ptr_vector<IntExceptionHandler> exceptions;

    /** Value stack. */
    IntDataSegment value_stack;

    /** Global context template. \see setNewGlobalContext */
    static std::auto_ptr<IntContext> global_context;

    /** Notification message associated with this process. */
    IntNotificationMessage* notification_message;

    /** Process kind. This is used for labelling and finding the process;
        it has no effect upon its execution. Think of this as an extension
        of the process name. (PCC 1.x actually used process names to find
        auto tasks, but this fails here since we do i18n.) */
    ProcessKind process_kind;

    /** Context top-of-stack. Registers the number of contexts that were on the
        process's context stack when it was created. This is used to identify
        the process it was invoked from. */
    uint16_t context_tos;
};


/** Push a new context. This one will be the first one used to look up variables. */
inline void
IntExecutionContext::pushNewContext(IntContext* ctx)
{
    contexts.push_back(ctx);
}

/** Pop context. Cancels a previous pushNewContext. */
inline void
IntExecutionContext::popContext()
{
    contexts.pop_back();
}

inline void
IntExecutionContext::markContextTOS()
{
    this->context_tos = contexts.size();
}

inline void
IntExecutionContext::setState(ProcessState ps)
{
    this->process_state = ps;
}

inline IntExecutionContext::ProcessState
IntExecutionContext::getState() const
{
    return process_state;
}

inline void
IntExecutionContext::setProcessKind(ProcessKind k)
{
    this->process_kind = k;
}

inline IntExecutionContext::ProcessKind
IntExecutionContext::getProcessKind() const
{
    return process_kind;
}

#endif
