/**
  *  \file int/vmio.h
  *  \brief Interpreter: Virtual Machine I/O
  *
  *  (c) 2001-2015 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_INT_VMIO_H
#define PCC_V2_INT_VMIO_H

#include <map>
#include <functional>
#include "int/bytecode.h"
#include "cpluslib/iterator.h"
#include "int/hash.h"
#include "int/array.h"
#include "cpluslib/ptrmap.h"
#include "int/subr.h"

class IntExecutionContext;
struct IntExecutionFrame;
class IntContext;
class IntStructureType;
class IntStructureTypeData;
class IntStructureValue;
class IntStructureValueData;

/** Context to save a virtual machine state subset.

    Users must call addBCO() and addProcess() to add objects to save. 
    They are actually saved using the save() method. We should avoid
    forward references in a VM file (although they are now allowed).
    The addXXX() pass therefore gathers objects to save in the correct
    order, while the save() pass just writes out what was gathered so
    far.

    Note that this stores a global instance pointer for use by
    serialisation functions; thus is is reentrant, but not
    thread-safe. */
class IntVMSaveContext {
    IntVMSaveContext* prev_instance;
    static IntVMSaveContext* save_instance;

    // Copy protection
    IntVMSaveContext(const IntVMSaveContext&);
    IntVMSaveContext& operator=(const IntVMSaveContext&);

    // Map of object to Id. The object can be anything (process, bytecode, array, hash).
    std::map<void*,uint32_t> obj_to_id;

    // Object Id counter. Contains the last Id assigned.
    uint32_t obj_id_counter;

    // Save plan.
    std::vector<void*> plan_objs;
    std::vector<uint8_t> plan_types;

    // Current process.
    IntExecutionContext* current_process;

    static void saveBCO(Stream& out, const IntBytecodeObject& bco, uint32_t id);
    static void saveHash(Stream& out, const IntHashData& hash, uint32_t id);
    static void saveArray(Stream& out, const IntArrayData& array, uint32_t id);
    static void saveStructureType(Stream& out, const IntStructureTypeData& type, uint32_t id);
    void saveStructureValue(Stream& out, const IntStructureValueData& value, uint32_t id);

    void saveFrame(Stream& out, const IntExecutionFrame& fr);
    void saveProcess(Stream& out, IntExecutionContext& exc);
 public:
    IntVMSaveContext();
    virtual ~IntVMSaveContext();

    // see int/vmioinst.cc for why these are virtual
    virtual uint32_t addBCO(IntBytecodeObject& bco);
    virtual uint32_t addHash(IntHashData& hash);
    virtual uint32_t addArray(IntArrayData& array);
    virtual uint32_t addStructureType(IntStructureTypeData& type);
    virtual uint32_t addStructureValue(IntStructureValueData& value);
    void     addProcess(IntExecutionContext& exc);
    void     save(Stream& out);

    IntExecutionContext* getCurrentProcess() const;

    static IntVMSaveContext* getCurrentInstance();
};

/** Context to load a virtual machine state subset.

    Note that this stores a global instance pointer for use by
    serialisation functions; thus is is reentrant, but not
    thread-safe. */
class IntVMLoadContext {
    class ObjectLoader;

    // Copy protection.
    IntVMLoadContext(const IntVMLoadContext&);
    IntVMLoadContext& operator=(const IntVMLoadContext&);

    // Previous instance.
    IntVMLoadContext* prev_instance;
    static IntVMLoadContext* load_instance;

    // Loaded objects. Indexed by object Id.
    ptr_map<uint32_t,IntSubroutineValue> bco_map;
    ptr_map<uint32_t,IntHash> hash_map;
    ptr_map<uint32_t,IntArray> array_map;
    ptr_map<uint32_t,IntStructureValue> struct_value_map;
    ptr_map<uint32_t,IntStructureType> struct_type_map;

    // Current process. Valid during loadProcess().
    IntExecutionContext* current_process;

    void loadBCO(ObjectLoader& ldr, uint32_t id);
    void loadHash(ObjectLoader& ldr, uint32_t id);
    void loadArray(ObjectLoader& ldr, uint32_t id);
    void loadStructureValue(ObjectLoader& ldr, uint32_t id);
    void loadStructureType(ObjectLoader& ldr, uint32_t id);
    void loadProcess(ObjectLoader&);
    void loadFrames(IntExecutionContext&, Stream& s, uint32_t count);
 public:
    IntVMLoadContext();
    ~IntVMLoadContext();

    IntBCORef getBCO(uint32_t id);
    Ptr<IntHashData> getHash(uint32_t id);
    Ptr<IntArrayData> getArray(uint32_t id);
    Ptr<IntStructureValueData> getStructureValue(uint32_t id);
    Ptr<IntStructureTypeData> getStructureType(uint32_t id);
    IntContext* makeFrameContext(uint32_t id) const;

    void load(Stream& s, bool acceptProcesses);
    IntExecutionContext* getCurrentProcess() const;

    static IntVMLoadContext* getCurrentInstance();
};

/** Get current process.
    \return Current process when we're in a addProcess/saveProcess call, 0 if none. */
inline IntExecutionContext*
IntVMSaveContext::getCurrentProcess() const
{
    return current_process;
}

/** Get current process.
    \return Current process when we're loading a process, 0 if none. */
inline IntExecutionContext*
IntVMLoadContext::getCurrentProcess() const
{
    return current_process;
}

#endif
