/**
  *  \file tools/c2mkturn.cc
  *  \brief "Maketurn" utility - Main Function
  *
  *  (c) 2001-2015 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include "cpluslib/format.h"
#include "game/gamedirs.h"
#include "game/maketurn.h"
#include "game/racename.h"
#include "game/reg.h"
#include "util/console.h"
#include "util/except.h"
#include "util/modinit.h"
#include "util/nls.h"
#include "version.h"
#include "arch/platform.h"
#include "game/storage/overview.h"

#define VERSION PCC2_VERSION

static const char* progname;

/** Exit with error. */
static void
die(string_t str)
{
    console.write(LOG_ERROR, format("%s: %s", progname, str));
    std::exit(1);
}

/** Exit with help message. */
static void
help()
{
    console << format(_("PCC2 Turn File Compiler v%s - (c) 2010-2015 Stefan Reuther\n"), VERSION)
            << format(_("\n"
                        "Usage:\n"
                        "  %s [-h]\n"
                        "  %$0s [gamedir [rootdir]]\n\n"
                        "Options:\n"
                        "  tbd.\n"
                        "\n"
                        "Report bugs to <Streu@gmx.de>\n"),
                      progname);
    console.flush();
    std::exit(0);
}

static const ModuleInitializer*const mkturn_modules[] = {
    mod_base_console,
    mod_reg,
    mod_racenames,
    0
};

int
main(int /*argc*/, char** argv)
{
    /* Arguments */
    const char* gameDir = 0;
    const char* rootDir = 0;
    bool optForce = false;

    /* Initialize */
    performInitialisation(OnStartup, mkturn_modules);
    argv = Platform::convertArgumentsToUtf8(argv);
    progname = argv[0];

    /* Parse */
    while (const char* arg = *++argv) {
        if (*arg == '-') {
            while (char c = *++arg) {
                switch (c) {
                 case 'h':
                    help();
                    break;
                 case 'f':
                    optForce = true;
                    break;
                 default:
                    die(format(_("invalid option specified. Use `%s -h' for help"), progname));
                    break;
                }
            }
        } else if (!gameDir) {
            gameDir = arg;
        } else if (!rootDir) {
            rootDir = arg;
        } else {
            die(_("too many arguments"));
        }
    }

    /* Set up game */
    if (!gameDir)
        gameDir = ".";
    initDirectories(gameDir, rootDir ? string_t(rootDir) : Platform::getDataDirectory());

    try {
        performInitialisation(OnEnterDirectory, mkturn_modules);
    }
    catch (FileProblemException& e) {
        die(format("%s: %s", e.getFileName(), e.what()));
    }
    catch (std::exception& e) {
        die(e.what());
    }

    /* Check */
    GDirectoryOverview ov;
    ov.scan(*game_file_dir);
    if ((ov.getGlobalFlags() & ov.have_Gamedir) == 0) {
        die(format(_("directory '%s' does not contain unpacked game data"), gameDir));
    }
    if ((ov.getGlobalFlags() & ov.have_Conflict) != 0 && !optForce) {
        die(format(_("directory '%s' contains data from different games.\n"
                     "NOTE: use '-f' to force compilation of turn files anyway"), gameDir));
    }

    GMaketurn theMaketurn(*game_file_dir);
    for (int pl = 1; pl <= NUM_PLAYERS; ++pl) {
        if (ov.getPlayerFlags(pl) & ov.have_Gamedir) {
            try {
                int ncommands = theMaketurn.makeTurn(pl);
                console.write(LOG_INFO, format(_("%s: %d command%!1{s%}"), player_racenames.getShortName(pl), ncommands));
            }
            catch (std::exception& e) {
                die(e.what());
            }
        }
    }

    /* Write them out */
    console.write(LOG_INFO, format(_("Writing %d turn file%!1{s%}..."), theMaketurn.getNumTurns()));
    theMaketurn.writeTurns();

    return 0;
}
